/* jshint esversion : 6 */
require('app-code-js');
const DBS_Connection = require('../db-connection/index')();
//
const { dataType, dataConverter: sqlDataConverter } = require('../../data-type');
const where = require('./where');


class MySqlQuery extends DBS_Connection {

    constructor(opt) {
        super(opt);
    }

    query() {

        const q = $q.defer();
        const args = Array.from(arguments);
        const lastArg = args[args.length - 1];

        const { _connectionParams: { connectionType } } = this._db;

        const __setQueryParams = connection => {

            const __defaultQueryCallback = (err, res, fields) => {
                connectionType === 'pool' ? connection.release() : connection.end();

                if (err) return q.reject(err);

                q.resolve(res);
            };

            const fn = (typeof lastArg === "function") ? args.pop() : __defaultQueryCallback;

            if (typeof lastArg !== "function")
                args.push(fn);

            return args;
        };

        this._db.connect().then(Connection => {
            if (connectionType === 'pool') {


                Connection.getConnection(function(err, connection) {
                    if (err)
                        return q.reject(err);

                    connection.query.apply(connection, __setQueryParams(connection));
                });
            } else {
                Connection.query.apply(connection, __setQueryParams(Connection));
            }
        });

        return q.promise;
    }


    /**
     * @param {string} tableName 
     * @param {<object>} opt {fields:(string|string[]), where:<object>}   
     */
    find(tableName, opt) {
        if (!opt)
            opt = {};

        const { fields = '*', db: dbName, where: whereOpt } = opt;

        let query = "SELECT ";

        if (!fields || 'string' === typeof fields && ['*', 'any'].indexOf(fields) > -1)
            query += '*';
        else if (Array.isArray(fields))
            query += fields.join();


        query += (` FROM ${this.db_dot_table(tableName, dbName)} ${where(whereOpt)} `);

        return this.query(query);
    }


    /**
     * 
     * @param {string} tableName 
     * @param {<object>} param1 {interval, autostop, {where:<object>}} 
     */
    watchQueryResult(tableName, { interval = 200, autostop = false, ...optFind }) {

        const that = this;

        let _watcherFn = null;

        const watcher = {

            _queryValue: [],

            start() {

                if (!_watcherFn) {
                    _watcherFn = function _wFn() {

                        that.find(tableName, optFind).then(
                            res => {
                                this._queryValue = res;

                                if (_watcherFn)
                                    setTimeout(_watcherFn, interval);

                            },
                            err => {
                                this.stop();
                                console.log(err);
                            }
                        );

                    }.bind(this);
                }

                _watcherFn();
            },

            stop() {
                if (_watcherFn)
                    _watcherFn = null;
            },

            onChange(fn) {
                this.watchProperty('_queryValue', (o, n) => {
                    fn(o, n);

                    if (this.autostop)
                        this.stop();
                });
            }
        };

        watcher.start();

        return watcher;
    }




    /**
     * @param {string} tableName 
     * @param {stirng} dbName 
     */
    prepareQuery(tableName, dbName) {

        const q = $q.defer();
        const showColumnsDefer = $q.defer();
        const indexDefer = $q.defer();
        const __promises = [showColumnsDefer.promise, indexDefer.promise];

        const r = {};


        this.showColumns(tableName).then(
            c => {
                const columns = c.map(e => e.toLowerKeys());

                r.columns = {
                    columns: columns.map(e => e.toLowerKeys()).filter(e => {
                        if (!e.extra.match(/auto_increment/i))
                            return e;
                    }),
                    get columnsName() {
                        return this.columns.map(e => e.field);
                    },
                    // primari key columns
                    pkColumns: columns.find(e => (e.key.match(/PRI/i) && e.extra.match(/auto_increment/i)))
                };
                //
                showColumnsDefer.notify(r.columns);
                showColumnsDefer.resolve();
            },
            showColumnsDefer.reject
        );


        this.groupIndexByName(tableName).then(
            c => {
                r.indexes = {
                    indexes: c.filter(e => {
                        if (!Array.isArray(e.columns))
                            e.columns = e.columns.replace(/\[|]/g, '').split(/,/g);

                        if (!e.primary)
                            return e;
                    }),
                    get indexedColumns() {
                        return this.indexes.map(e => e.columns).reduce((curr, next) => curr.concat(next), []).unique();
                    },
                    get IndexdeColumnsLength() {
                        return this.indexedColumns.length;
                    }
                };
                //
                indexDefer.notify(r.indexes);
                indexDefer.resolve();
            },
            indexDefer.reject
        );

        $q.all(__promises).then(() => q.resolve(r), q.reject, q.notify);

        return q.promise;
    }


    /**
     * @param {string} tableName "name of table" 
     * @param {<object>} a "array of data to  update" 
     * @param {<object>} {dbName:string, where:<object>}
     */
    insert(tableName, a, opt) {

        const q = $q.defer();

        this.prepareQuery(tableName).then(
            c => {

                const { db: dbName, where: whereOpt } = (opt || {});
                //
                const { columns: { columns, columnsName } } = c;

                const dataConverter = sqlDataConverter(columns, a);

                const data = dataConverter.getConvertedData();


                const { dbms, version } = this._db;
                //
                const query = cmpInsert({
                    tableName: this.db_dot_table(tableName, dbName),
                    columnsName: columnsName,
                    data: data,
                    cmd: 'insert',
                    dbms: dbms,
                    version: version
                }) + where(whereOpt);

                this.query(query).then(q.resolve, q.reject);

            },
            q.reject,
            q.notify
        );

        return q.promise;
    }


    /**
     * @param {string} tableName "name of table" 
     * @param {<object>} a "array of data to  update" 
     * @param {<object>} {dbName:string, where:<object>}
     */
    update(tableName, a, opt) {

        const q = $q.defer();

        this.prepareQuery(tableName).then(
            c => {

                const { db: dbName, where: whereOpt } = (opt || {});
                //
                const { columns: { columns, columnsName } } = c;

                const dataConverter = sqlDataConverter(columns, a);
                const data = dataConverter.getConvertedData();

                //

                let query = `UPDATE ${this.db_dot_table(tableName, dbName)} SET `;

                query += data.map(e => Object.entries(e).map(x => x.join('=')).join())[0];
                query += where(whereOpt);



                //
                //
                this.query(query).then(q.resolve, q.reject);

            },
            q.reject,
            q.notify
        );

        return q.promise;
    }

    /**
     * @param {string} tableName "name of table" 
     * @param {<object>} a "array of data to insert or update" 
     * @param {string} dbName "name of db"
     */
    put(tableName, a, dbName) {

        const q = $q.defer();

        this.prepareQuery(tableName).then(
            c => {

                const { columns: { columns, columnsName }, indexes: { indexedColumns, IndexdeColumnsLength } } = c;
                //
                const dataConverter = sqlDataConverter(columns, a);
                const data = dataConverter.getConvertedData();
                const dataColumns = data.map(e => Object.keys(e))[0];


                const indexColumnsExist = () => {
                    const a = dataColumns.filter(e => {
                        if (indexedColumns.find(o => o == e))
                            return e;
                    });

                    return (a.length === IndexdeColumnsLength);
                };



                if (indexColumnsExist()) {

                    const { dbms, version } = this._db;
                    //
                    let query = cmpInsert({
                        tableName: this.db_dot_table(tableName, dbName),
                        columnsName: columnsName,
                        data: data,
                        cmd: 'insert',
                        dbms: dbms,
                        version: version
                    }) + ' ON DUPLICATE KEY UPDATE ';

                    const __value__ = (dbms === 'mariadb' && version > '10.3.2') ? 'VALUE' : 'VALUES';
                    query += columnsName.map(e => `${e} = ${__value__}(${e})`).join();

                    this.query(query).then(q.resolve, q.reject);

                } else {
                    q.reject("all elements in a[{}] must have keys " + indexColumns.join() + '\nyou insert alemente with those keys : ' + dataColumns.join());
                }

            },
            q.reject,
            q.notify
        );

        return q.promise;
    }


    delete(tableName, { where: whereOpt }, dbName) {
        let query = 'DELETE FROM ' + this.db_dot_table(tableName, dbName) + where(whereOpt);
        return this.query(query);
    }


    db_dot_table(tableName, dbName) {
        const db = dbName || this.dbName;
        return (db + '.' + tableName);
    }

}


/**
 * @param {string} tableName "name of table" 
 * @param {string[]} columnsName "list of table columns" 
 * @param {<object>} data "data to insert" 
 */
function cmpInsert({ tableName, columnsName, data, cmd, dbms, version }) {

    let query = "INSERT INTO " + tableName;

    if (!Array.isArray(data))
        data = [data];

    const __value__ = (dbms === 'mariadb' && version > '10.3.2') ? 'VALUE' : 'VALUES';

    if (cmd === 'update')
        query += ` (${columnsName.join()}) ${__value__} `;
    else {
        const cc = Object.keys(data[0]).filter(e => {
            if (columnsName.indexOf(e) > -1)
                return e;
        });
        query += ` (${cc.join()}) ${__value__} `;
    }

    query += data.map(e => {
        const o = {};

        const a = Object.keys(e);

        columnsName.forEach(f => {
            if (cmd == 'update')
                o[f] = (a.indexOf(f) === -1) ? ` ${__value__}(${f})` : e[f];
            else if (a.indexOf(f) > -1)
                o[f] = e[f];
        });

        return `(${Object.values(o)})`;

    }).join();



    return query;
}


module.exports = MySqlQuery;